/*
 * Null garbage collector.  Simply allocates linearly until the heap fills up,
 * then dies with an out of memory error.  Good for some kinds of testing.
 * 
 * Author: Robin Garner
 */
 
#include "object_model.h"
#include "vm_gc.h"
#include "gc_vm.h"

#define NULLPTR ((Address)0)

/*
 * The bounds of the heap regions.
 */
static Address IMMORTAL, IMMORTAL_LIMIT;

/*
 * Active heap pointers
 */
static Address immortal;

/*
 * Initialize the heap
 */
void initialize(int minSize, int maxSize) {
 IMMORTAL = malloc(maxSize);
 IMMORTAL_LIMIT = IMMORTAL + maxSize;
}

/*
 * Allocate space for an object
 */
Address alloc(int bytes, int allocator) {
  Address result = bp_alloc(bytes,&immortal,IMMORTAL_LIMIT);
  if (result == NULL) {
    outOfMemory("immortal",bytes);
  }
  return result;
}

/*
 * General bump-pointer allocation
 * Returns null if the limit is exceeded.
 */
__inline__ 
static Address bp_alloc(int bytes, Address *bp, Address limit) {
  Address result = *bp;
  *bp += bytes;
  if( *bp >= limit )
    return NULL;
  else
    return result;
}

static void outOfMemory(char *space, int bytes) {
  fprintf(stderr,"Heap region \"%s\" full, last request %d bytes\n",space,bytes);
  exit(2);
}
		      
/*
 * Collect - no-op
 */
void collect() {
  // Do nothing
}

void addRootLocation(Address slot) { }

/*
 * For performance we would inline this in the header file
 */
ObjectReference readBarrier(ObjectReference obj, Offset offset, int event) {
	return *(ObjectReference *)((Address)obj + offset));
}

/*
 * For performance we would inline this in the header file
 */
void writeBarrier(ObjectReference obj, Offset offset, 
		  ObjectReference newVal, int event) {
    *(ObjectReference *)((Address)obj + offset)) = newVal;
}
