/*
 * Simple semi-space collector
 *
 * Implements an uncollected immortal space, too.
 *
 * Initial implementation is very naive in its management of real memory -
 * we use malloc to allocate the heap.  This should become 'mmap' in time.
 */
 
#include "object_model.h"
#include "vm_gc.h"
#include "gc_vm.h"


#define TRUE (1 == 1)
#define FALSE (!TRUE)

/*
 * GC bit mask
 */
#define BITMASK_COPIED  = 1    // Object has been copied, forwarding word is valid
#define BITMASK_COPYING = 2    // Object in the process of being copied (for future par. collector)
#define BITMASK_ALL     = 3    // All the GC bits

/*
 * The bounds of the heap regions.
 */
static Address IMMORTAL, IMMORTAL_LIMIT;
static Address SS[2], SS_LIMIT[2];

/*
 * Active heap pointers
 */
static Address immortal;
static Address ss, ss_limit;  // Putting these in registers should speed it up
static int toSpace = 0, fromSpace = 1;

/*
 * Switch from and to spaces
 */
void flip() {
  toSpace = (toSpace + 1) % 2;
  fromSpace = 1 - toSpace;
  ss = SS[toSpace];
  ss_limit = SS_LIMIT[toSpace];
}


/*
 * Helper functions for debugging-style assertions 
 */
int inImmortalSpace(Address ptr) {
  return IMMORTAL <= ptr && ptr < IMMORTAL_LIMIT;
}

int inToSpace(Address ptr) {
  return SS[toSpace] <= ptr && ptr < SS_LIMIT[toSpace];
}

int inFromSpace(Address ptr) {
  return SS[fromSpace] <= ptr && ptr < SS_LIMIT[fromSpace];
}

/*
 * Initialize the heap
 */
void initialize(int minSize, int maxSize) {
	int immortalSize = maxSize / 10;
	int halfSize = (maxSize - immortalSize)/2;
	int i;
	
	IMMORTAL = malloc(immortalSize);
	IMMORTAL_LIMIT = IMMORTAL + immortalSize;
	for( i=0; i < 2; i++ ) {
	    SS[i] = malloc(halfSize);
	    SS_LIMIT[i] = SS[i] + halfSize;
	}
	flip();  // Sets the volatile heap pointers as a side effect
	init_queue();  // Initialize the scanning queue
}


static void outOfMemory(char *space, int bytes) {
  fprintf(stderr,"Heap region \"%s\" full, last request %d bytes\n",space,bytes);
  exit(2);
}

/*************************************************************************
 * 
 * Allocate space for an object
 * 
 */
Address alloc(int bytes, int allocator) {
  switch (allocator) {
  case ALLOC_IMMORTAL: 
    return alloc_immortal(bytes,align);

  default:
    return alloc_ss(bytes,align);
  }
}

/*
 * General bump-pointer allocation
 * Returns null if the limit is exceeded.
 */
__inline__ 
static Address bp_alloc(int bytes, Address *bp, Address limit) {
  Address result = *bp;
  *bp += bytes;
  if( *bp >= limit )
    return NULL;
  else
    return result;
}

/*
 * Allocate space from the immortal space
 */
Address alloc_immortal(int bytes) {
  Address result = bp_alloc(bytes,&immortal,IMMORTAL_LIMIT);
  if (result == NULL) {
    outOfMemory("immortal",bytes);
  }
  return result;
}
		      
/*
 * Allocate space in the semispace
 */
Address alloc_ss(int bytes) {
  return alloc_ss(bytes,align,FALSE);
}

__inline__ 
static Address alloc_ss(int bytes, int justCollected) {
  Addressresult = bp_alloc(bytes,&ss,ss_limit);
  if (result == NULL) {
    if (justCollected) {
      outOfMemory("semi-space",bytes);
    }
    gc();
    return alloc_ss(bytes,align,TRUE);
  }
  return result;
}

/***************************************************************************
 * 
 * Garbage collection
 */

/*******************************************
 * Queue of objects waiting to be scanned
 * 
 * This should be abstracted out as a proper 'queue' type.
 */
#define INITIAL_QUEUE_SIZE = 1024;

ObjectReference queue[];
int queue_size = 0;
int queue_size_hwm = INITIAL_QUEUE_SIZE;
int queue_max = -1;

static void init_queue() {
	queue_size = queue_size_hwm == 0 ? INITIAL_QUEUE_SIZE : queue_size_hwm;
	queue = (ObjectReference[]) malloc(queue_size*sizeof(ObjectReference));
}

static void grow_queue() {
	int new_queue_size = queue_size * 2;
	ObjectReference[] newQueue = (ObjectReference[]) malloc(new_queue_size*sizeof(ObjectReference));
	portable_memcpy(newQueue,queue,queue_size*sizeof(ObjectReference));
	free(queue);
	queue = newQueue;
}

static int isEmpty() {
	return queue_max == -1;
}

static void push(ObjectReference object) {
	if( queue_max >= queue_size ) grow_queue();
	queue[++queue_max] = object;
}

static ObjectReference pop() {
	if( isEmpty() ) {
		fprintf(stderr,"Pop from empty queue - THIS CAN'T HAPPEN\n");
		exit(1);
	}
	return queue[queue_max--];
}

/*
 * Forward an object
 */
ObjectReference forward(ObjectReference object) {
	Address objectAddress = objectReferenceToAddress(object);
	struct object_header *header = (struct object_header *)objectAddress;
	
	/* Has the object already been copied ? */
	if( !(header->gcbits & BITMASK_COPIED) ) {
		/* Allocate space for the copied object */
		Address newObjectAddress = alloc_ss(getSizeWhenCopied(object));
		copyObject(newObjectAddress,objectAddress);
		
		/* Install the forwarding pointer and update the object metadata */
		header->fwptr = addressToObjectReference(newObjectAddress);
		push(header->fwptr);
	}
	return header->fwptr;
}

/*
 * Scan an object during GC.  Forward all its reference fields.
 */
static void scan(ObjectReference object) {
    int fields, offsets[], i;
    
    /* Useful optimization - 30% of objects have no pointers */
    if( hasPointers(object) ) return;
    
    /* Forward the objects this one points to */
    if( isArray(object) ) {
        ObjectReference *slots[] = (ObjectReference **)object;
        for( i=0; i < header->length; i++ ) {
            if( *slots[i] != NULL )
                *slots[i] = forward(*slots[i]);
        }
    } else {
        fields = referenceFieldCount(object);
        offsets = referenceFieldOffsets(object);
    
        for( i=0; i < fields; i++ ) {
            ObjectReference *slot = (ObjectReference *)((Address)object + offsets[i]);
            if( *slot != NULL )
                *slot = forward(*slot);
        }
    }   
}

/*
 * Collect
 *
 * 1. Flip from-space and to-space.  
 * 2. Enumerate roots, and copy the root-live
 */
static void gc() {
		
    /* Flip */
	flip();
		  
	/* VM enumerates the roots */
	enumerateRoots();
	
	while( !isEmpty() ) {
        ObjectReference obj = pop();
        scan(obj);
    }
}


/*
 * Collect - externally visible.
 */
void collect() {
  gc();
}

/*
 * Enumerate a root location during a GC.
 * 
 * Forward the object.
 */
void addRootLocation(Address* slot) {
    if( *slot != NULL )
	    *slot = forward(*slot);
}

/*
 * For performance we would inline this in the header file
 * 
 * In a semispace collector this is a no-op.
 */
ObjectReference readBarrier(ObjectReference obj, Offset offset, int event) {
	return *(ObjectReference *)((char *)obj + offset));
}

/*
 * For performance we would inline this in the header file
 * 
 * In a semispace collector this is a no-op.
 */
void writeBarrier(ObjectReference obj, Offset offset, 
		  ObjectReference newVal, int event) {
    *(ObjectReference *)((char *)obj + offset)) = newVal;
}
