#ifndef _VM_GC_H
#define _VM_GC_H

#include "object_model.h"

/*
 * The VM's interface to the GC.
 *
 * Minimal
 *
 * @author Robin Garner
 */

typedef char * Address;   // Pointer to a single-byte type, makes address arithmetic easier

typedef int Offset;

/*
 * Initialize the heap
 */
void initialize(int minBytes, int maxBytes);

/*
 * Define constants for 'allocators', which provide hints to the GC
 * as to how to allocate the object
 */

/* Default allocator - let the GC decide */
#define ALLOC_DEFAULT 0

/* Immortal objects - they don't die and they don't move */
#define ALLOC_IMMORTAL 1

/* Stack objects */
#define ALLOC_STACK 2

/* Large objects */
#define ALLOC_LARGE 3

/*
 * Allocate space for an object.
 */
Address alloc(int bytes, int allocator);

/*
 * Request collection 
 */
void collect();

/*
 * Add a root location during enumeration of roots
 *
 * 'slot' is the address of the pointer, not the address of the heap
 * object (otherwise moving collectors won't work).
 */
void addRootLocation(Address slot);

/*
 * Barriers.  Whenever the VM reads or writes a pointer to the heap it
 * must do so via these functions, so that the GC gets a chance to 
 * intercept the read or write.
 */

#define BARRIER_GETFIELD   1
#define BARRIER_PUTFIELD   2
#define BARRIER_GETSTATIC  3
#define BARRIER_PUTSTATIC  4
#define BARRIER_AASTORE    6
#define BARRIER_AALOAD     5

/*
 * Read barrier.  Read an object reference from the given offset into 
 * the given object.  Return the reference.
 */
ObjectReference readBarrier(ObjectReference obj, Offset offset, int event);

/*
 * Write barrier.  Write the value 'newVal' to the given offset into the given 
 * object.
 */
void writeBarrier(ObjectReference obj, Offset offset, 
		  ObjectReference newVal, int event);

#endif  // conditional include
